#!/usr/bin/env python3

"""
 " These need to be updated with each release.
"""

LINARO_RELEASE              = "18.01"

JUNO64_ANDROID_RELEASE      = "18.01"
JUNO64_OE_RELEASE           = "17.01"
JUNO64_OE_MINI_VSN          = "5.2_20170127-761"
JUNO64_OE_LAMP_VSN          = "5.2_20170127-761"
JUNO64_PREBUILT_RELEASE     = "18.01"

JUNO32_OE_RELEASE           = "15.07"
JUNO32_OE_ALIP_VSN          = "4.9_20150725-725"
JUNO32_PREBUILT_RELEASE     = "18.01"

FVP64_ANDROID_RELEASE       = "7.0-16.10"
FVP64_OE_RELEASE            = "15.09"
FVP64_OE_MINI_VSN           = "4.9_20150912-729"
FVP64_OE_LAMP_VSN           = "4.9_20150912-729"
FVP64_PREBUILT_RELEASE      = "18.01"

FVP32_OE_RELEASE            = "15.07"
FVP32_OE_ALIP_VSN           = "4.9_20150725-725"
FVP32_PREBUILT_RELEASE      = "18.01"

TC2_ANDROID_RELEASE         = "6.0-15.11"
TC2_OE_RELEASE              = "15.07"
TC2_OE_ALIP_VSN             = "4.9_20150725-725"

COMPILER_RELEASE            = "6.2-2016.11"
AARCH32_COMPILER_VSN        = "6.2.1-2016.11"
AARCH64_COMPILER_VSN        = "6.2.1-2016.11"

SCRIPT_DEPS = [ "bison", "flex", "g++", "git", "libssl-dev",
                "python-crypto", "python-wand", "uuid-dev" ]

"""
 " Dictionary keys used to uniquely identify Resources and Filesystems.
"""

KEY_OE_MINI_IMAGE           = 0xefa8d8955fc34af791b82cef0b6cb774
KEY_OE_LAMP_IMAGE           = 0x37052d5201fc4617af81c8e2c5805a57
KEY_OE_ALIP_IMAGE           = 0xec810b6bf9a9446ab997592018a71391
KEY_ANDROID_IMAGE           = 0xdae105d10fd148f99a4c34ac190d7d6d
KEY_ANDROID_RAMDISK         = 0xb9e1090c529b4178ae1bdaf2d3a244e7

KEY_BUSYBOX_LATEST_PREBUILT = 0xf22252ec495a457587ca1c5c749935eb
KEY_BUSYBOX_ACK_PREBUILT    = 0xf200237c35b44951bdbf7bdcbf679a19
KEY_OE_PREBUILT             = 0xfa937ff74ca4460eb0523f219a69b9b6
KEY_OE_DEBUG_PREBUILT       = 0xd60e0bf17d60453e9cbaf188856c05da
KEY_ANDROID_PREBUILT        = 0x8ad52ef7d91045abbdabeee3fab91834
KEY_ANDROID_DEBUG_PREBUILT  = 0x5a25efab1ce744a99967a44b07ce17cf
KEY_UEFI_PREBUILT           = 0x7df1803e89b946be923965602e0916f4

KEY_BUSYBOX_FS              = 0x89135a7a281942318b18aa88dfd62d23
KEY_OE_MINI_FS              = 0x1e14bd11d4734e95a649c85b0edab9ca
KEY_OE_LAMP_FS              = 0xe0cfab2306b04e28ba18aeae66168feb
KEY_OE_ALIP_FS              = 0x1cf91f101842495b80ff6c9806a6c4ee
KEY_ANDROID_FS              = 0x53888ea21e9248c08dcc71355ec2fbe2

KEY_AARCH32_COMPILER        = 0x59a3802e2da7477680dadd2a399b520e
KEY_AARCH64_COMPILER        = 0x262a290a26da4fecbb31be96218ed3ab

import bz2
import ctypes
import hashlib
import getopt, gzip
import inspect
import math
import os
import platform
import re
import shutil, stat, subprocess, sys
import tarfile, traceback, types
import urllib.error, urllib.request
import zipfile

from collections import namedtuple
from functools import reduce

"""
 " Script state and configuration.
"""
class Script():

    __basedir = None
    __dldir = None
    __verbose = False
    __debug = False
    __testmode = False

    """
     " All file operations are performed relative to the script's "base"
     " directory, which is either the script's resident directory or the
     " working directory from which the script was invoked.
    """
    def BaseDir():
        if Script.__basedir is None:
            scriptdir = sys.path[0]+"/"
            workdir = os.getcwd()+"/"
            Script.__basedir = workdir if (workdir == scriptdir) else \
                workdir if 0 == Menu.Prompt\
                (
                    title="[WARNING] Current working directory differs from "
                    "script resident directory.\nPlease specify which directory"
                    " to initialise as your workspace:",
                    choices=\
                    [
                        Menu.Choice(workdir, "Working directory"),
                        Menu.Choice(scriptdir, "Script directory"),
                    ],
                    ret="index",
                ) else scriptdir
            try:
                os.chdir(Script.__basedir)
            except OSError as e:
                Script.Abort("Failed to cd from '"+workdir+"' to '"+
                    Script.__basedir+"'\n\nOSError: "+str(e))
        return Script.__basedir

    """
     " All downloads are stored in a hidden `.downloads/' directory.
    """
    def DlDir():
        if Script.__dldir is None:
            Script.__dldir = Shell.MkDir("downloads", hidden=True)+"/"
        return Script.__dldir

    def SetVerbose():
        Script.__verbose = True

    def Verbose():
        return Script.__verbose

    def SetDebug():
        Script.__debug = True
        Script.SetVerbose()

    def Debug():
        return Script.__debug

    def EnableTestMode():
        Script.__testmode = True

    def TestModeEnabled():
        return Script.__testmode

    def RunningOnLinux():
        return platform.system() == "Linux"

    def RunningOnWindows():
        return platform.system() == "Windows"

    """
     " Log entry to a function, printing the function name and its arguments.
     " Ignored unless Script.Debug() is True.
    """
    def BeginFunc():
        if Script.Debug():
            (frame, _, line, fnn, _, _) = \
                inspect.getouterframes(inspect.currentframe())[1]
            (args, _, _, locals) = inspect.getargvalues(frame)
            print("Calling "+fnn+"() @ line "+str(line))
            for arg in args:
                print("¬ "+arg+" = "+str(locals[arg]))

    """
     " Log a message prefixed with the calling function's name.
     " Ignored unless Script.Verbose() is True.
    """
    def Log( msg ):
        if Script.Verbose():
            (_, _, _, fnn, _, _) = \
                inspect.getouterframes(inspect.currentframe())[1]
            print(fnn+"(): "+msg)

    def Abort( msg ):
        (_, _, _, fnn, _, _) = \
            inspect.getouterframes(inspect.currentframe())[1]
        print("\n\n[FATAL] "+fnn+"(): "+msg+"\n")
        if not Script.TestModeEnabled():
            exit(1)

    """
     " Parse command-line arguments and initialise the script.
    """
    def Init():

        def usage():
            print("\nUsage:\n\n    "+__file__.rsplit("/")[-1]+" [opts]")
            print("\nWhere [opts] is any number of:\n")
            print("    `--verbose'      Print verbose logging information.\n")
            print("    `--debug'        Log function calls along with their"\
                +" arguments.\n"+(21*" ")+"Implies `--verbose'.\n")
            print("    `--testurls'     Test online resource download links.\n"\
                +(21*" ")+"Intended for internal ARM testing.\n")
            sys.exit(1)

        try:
            opts, _ = getopt.getopt(sys.argv[1:], "", \
                ["verbose", "debug", "testurls", "help"])
            for o, a in opts:
                if o in "--verbose":
                    Script.SetVerbose()
                elif o in "--debug":
                    Script.SetDebug()
                elif o in "--testurls":
                    Script.EnableTestMode()
                elif o in "--help":
                    usage()
        except getopt.GetoptError:
            usage()

        # First call sets base directory.
        Script.BaseDir()

"""
 " Shell utilities (file operations etc).
"""
class Shell():

    """
     " Create a new (optionally hidden) directory relative to Script.BaseDir().
    """
    def MkDir( path, hidden=False ):
        Script.BeginFunc()
        path = path.lstrip("/").rstrip("/")
        if hidden:
            # Strip away the directory tree to get the filename, then prefix it.
            # with "."
            filename = path.split("/")[-1]
            path = path.rstrip(filename)+"."+filename
            Script.Log("New path: "+path)
        try:
            fullpath = Script.BaseDir()+path
            if not os.path.isdir(fullpath):
                if os.path.isfile(fullpath):
                    raise OSError("Non-dir file exists with same name")
                Script.Log("Creating directory")
                os.makedirs(fullpath)
            if hidden and Script.RunningOnWindows():
                # 0x2 == FILE_ATTRIBUTE_HIDDEN (see MSDN `SetFileAttributes').
                Script.Log("Calling kernel32 to hide directory")
                ret = ctypes.windll.kernel32.SetFileAttributesW(fullpath, 0x2)
                if 0 == ret:
                    raise ctypes.WinError()
            return path
        except OSError as e:
            Script.Abort("Failed to create directory `"+fullpath+"'\n\n"+
                "OSError: "+str(e))

    """
     " Copy a file/directory relative to Script.BaseDir().
    """
    def Copy( src, dst ):
        Script.BeginFunc()
        src = Script.BaseDir()+src
        dst = Script.BaseDir()+dst
        try:
            shutil.copy(src, dst)
        except OSError as e:
            Script.Abort("Failed to copy `"+src+"' to `"+dst+"'\n\n"+
                "OSError: "+str(e))

    """
     " Move a file/directory relative to Script.BaseDir().
    """
    def Move(src, dst):
        Script.BeginFunc()
        src = Script.BaseDir()+src
        dst = Script.BaseDir()+dst
        try:
            shutil.move(src, dst)
        except OSError as e:
            Script.Abort("Failed to move `"+src+"' to `"+dst+"'\n\n"+
                "OSError: "+str(e))

    """
     " Remove a file relative to Script.BaseDir().
     " Note: Use RmDir() to remove a directory.
    """
    def RmFile( path ):
        Script.BeginFunc()
        path = Script.BaseDir()+path
        try:
            os.remove(path)
        except OSError as e:
            Script.Abort("Failed to remove file `"+path+"'\n\nOSError: "+str(e))

    """
     " Remove a directory relative to Script.BaseDir().
     " Note: Use RmFile() to remove a file.
    """
    def RmDir( path ):
        Script.BeginFunc()
        path = Script.BaseDir()+path
        try:
            shutil.rmtree(path)
        except OSError as e:
            Script.Abort("Failed to remove directory `"+path+"'\n\n"+
                "OSError: "+str(e))

    """
     " Set executable permission bits for user/group/others (i.e. `chmod a+x')
     " of a file relative to Script.BaseDir(). Ignored unless running on Linux.
    """
    def ChmodAX( path ):
        Script.BeginFunc()
        if not Script.RunningOnLinux():
            Script.Log("Function void when not running on Linux")
        else:
            path = Script.BaseDir()+path
            try:
                os.chmod(path, os.stat(path).st_mode | stat.S_IEXEC \
                    | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)
            except OSError as e:
                Script.Abort("Failed to `chmod a+x' file '"+path+"'\n\n"
                    "OSError: "+str(e))

    """
     " Erase the current console line.
    """
    def EraseLn():
        # On Linux assume we're running on a VT100 compatible terminal.
        if Script.RunningOnLinux():
            print("\r\033[K", end="")
        elif Script.RunningOnWindows():
            print("\r"+(" "*(shutil.get_terminal_size().columns-1))+"\r", end="")

    """
     " Flush stdout.
    """
    def Flush():
        sys.stdout.flush()

    """
     " Extract an archive relative to Script.BaseDir(). Returns True on
     " successful extraction, else False (example: unrecognised archive format).
    """
    def ExtractArchive( path, dstdir ):
        Script.BeginFunc()
        path = Script.BaseDir()+path
        dstdir = Script.BaseDir()+dstdir
        extracted = False
        try:
            # In Python some archive formats are exposed as binary file
            # wrappers that we can simply raw copy to another file descriptor.
            # We can handle all of these in the same way.
            binarywrapper = False

            if path.endswith(".tar.xz"):
                Script.Log("File is `.tar.xz' format")
                with tarfile.open(path, "r:xz", errorlevel=1) as f:
                    f.extractall(dstdir.rstrip(".tar.xz"))
                extracted = True

            elif path.endswith(".zip"):
                Script.Log("File is `.zip' format")
                with zipfile.ZipFile(path, "r") as f:
                    f.extractall(dstdir.rstrip(".zip"))
                extracted = True

            # ".bz2" is exposed as a binary file wrapper.
            elif path.endswith(".bz2") and not path.endswith(".tar.bz2"):
                Script.Log("File is `.bz2' format (binary wrapper bz2.BZ2File)")
                openfunc = bz2.BZ2File
                trun = -len(".bz2")
                binarywrapper = True

            # ".gz" is exposed as a binary file wrapper.
            elif path.endswith(".gz") and not path.endswith(".tar.gz"):
                Script.Log("File is `.gz' format (binary wrapper gzip.open)")
                openfunc = gzip.open
                trun = -len(".gz")
                binarywrapper = True

            # For binary file wrappers simply copy to new file descriptor.
            if binarywrapper and not extracted:
                Script.Log("Handling as binary wrapper")
                with openfunc(path, "rb") as inf:
                    with open(dstdir+path[:trun].rsplit("/")[-1], "wb") as outf:
                        while True:
                            chunk = inf.read(0x8000)
                            if not chunk:
                                break
                            outf.write(chunk)
                extracted = True

            if not extracted:
                Script.Log("No handler for supplied file format")

            return extracted
        except (tarfile.TarError, OSError) as e:
            Script.Abort("Failed to extract archive `"+path+"'\n\n"
                "TarError/OSError: "+str(e))

    class MD5():

        """
         " MD5 hash a file relative to Script.BaseDir().
        """
        def HashFile( path ):
            Script.BeginFunc()
            path = Script.BaseDir()+path
            md5hash = hashlib.md5()
            try:
                with open(path, "rb") as f:
                    while True:
                        block = f.read(md5hash.block_size)
                        if not block:
                            break
                        md5hash.update(block)
                return md5hash.hexdigest()
            except OSError as e:
                Script.Abort("Failed to MD5 hash file `"+path+"'\n\n"
                    "OSError: "+str(e))

        """
         " Exctract a hash from an MD5 sums file relative to Script.BaseDir().
        """
        def ExtractHash( sumspath, filename ):
            Script.BeginFunc()
            sumspath = Script.BaseDir()+sumspath
            md5 = None
            try:
                with open(sumspath, "r") as f:
                    # Match on standard MD5 checksum file formatted filename.
                    regex = re.search(r"(.+)\s\s"+filename+"[^\.]", f.read())
                    if not regex:
                        Script.Abort("Failed to find MD5 checksum for `"
                            +filename+"' in '"+sumspath+"'")
                    else:
                        md5 = regex.group(1)
                return md5
            except OSError as e:
                Script.Abort("Failed to extract MD5 checksum from file `"
                    +path+"'\n\nOSError: "+str(e))

"""
 " Downloading and manifesting utilities.
"""
class Web():

    """
     " Fetch a URL with pretty-printed progress. Returns True on success.
    """
    def FetchURL( url, dst, name, silent=False ):

        """
         " Format raw number of bytes into a 2dp precision SI suffixed string,
         " e.g. 2799698 --> "2.67 MiB".
        """
        def fmt( num ):
            if num <= 0:
                return "0 B"
            # Going to assume we won't be downloading files any bigger... :)
            suffixes = [ "B", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB"]
            log = int(math.floor(math.log(num, 1024)))
            sz = round(num / (1024 ** log), 2)
            return "{:.2f} {:}".format(sz, suffixes[log])

        Script.BeginFunc()
        dst = Script.BaseDir()+dst
        try:
            response = urllib.request.urlopen(url)
            # If server hosting URL supports 'Content-Length' header then we
            # can pretty-print the object's total size when reporting current
            # progress, e.g. "2.67 MiB / 1.1 GB"; if not we just pretty-print
            # the current running total e.g. "2.67 MiB".
            try:
                endln = " / "+\
                    fmt(int(response.getheader("Content-Length").strip()))
            except ValueError():
                Script.Log("Server does not support 'Content-Length' header")
                endln = ""

            # We can now read the URL as if it were a binary file.
            with open(dst, "wb") as f:
                progress = 0
                while True:
                    chunk = response.read(0x8000)
                    if not chunk:
                        break
                    f.write(chunk)
                    progress += len(chunk)
                    if (not silent) or Script.TestModeEnabled():
                        Shell.EraseLn()
                        print("Fetching "+name+": "+fmt(progress)+endln, end="")
                        Shell.Flush()

            return True
        except (OSError, urllib.error.HTTPError) as e:
            errtype = "OSError" if type(e) is OSError else "HTTPError"
            Script.Abort("Failed to download `"+name+"' from URL '"+url+"' to "
                "local file '"+dst+"'\n\n"+errtype+": "+str(e))
            # When Script.TestModeEnabled() is True, Script.Abort() won't exit,
            # so we need to indicate to the caller whether the download was
            # successul.
            return False

    """
     " Initialise and sync a manifest using repo.
    """
    def RepoSync( manifest ):

        Script.BeginFunc()
        if not Script.RunningOnLinux():
            Script.Log("Function void when not running on Linux")
            return

        # Silently fetch repo and make it executable.
        Script.Log("Fetching repo")
        Resource("Repo", "https://storage.googleapis.com/git-repo-downloads",
            "repo", dstdir="tools/").Fetch()
        repo = "tools/repo"
        Shell.ChmodAX(repo)

        # Pipe repo's stdout and stderr to /dev/null unless running verbose.
        if Script.Verbose():
            po, pe = None, None
        else:
            po, pe = subprocess.DEVNULL, subprocess.DEVNULL

        # Initialise repo.
        print("Initialising Repo tool... ", end="")
        Shell.Flush()
        if not 0 == subprocess.call\
        (
            [
                repo, "init", "-b", LINARO_RELEASE,
                "-m", "pinned-"+manifest+".xml", "-u",
                "https://git.linaro.org/landing-teams/working/arm/manifest",
            ],
            stdout=po, stderr=pe,
        ):
            Script.Abort("Failed to initialise Repo tool")
        print("Done.\nSyncing Repo tool (this may take a long time, please be"
            " patient)... ", end="")
        Shell.Flush()

        # Sync repo.
        if not 0 == subprocess.call([repo, "sync", "-j8"], stdout=po, stderr=pe):
            Script.Abort("Failed to sync Repo tool")
        print("Done.")

"""
 " Tool for presenting a menu to the user and asking them to make a selection.
"""
class Menu():

    """
     " Equivalent to class with constructor (name, descr=None, disabled=False).
    """
    Choice = namedtuple("MenuChoice", ["name", "descr", "disabled"])
    Choice.__new__.__defaults__ = None, False

    """
     " Prompt user to make a selection from a list of MenuChoice namedtuples.
     " Use `ret="index"' to return index of selection in choices list, or
     " use `ret="reference"' to return a reference to the choice itself.
    """
    def Prompt( title, choices, ret="index" ):

        """
         " Format a MenuChoice namedtuple into a printable string menu entry,
         " padding the name to a specified width before the description.
        """
        def mkentry( c, pad ):
            entry = "\n{:>2}) {:<{:}}".format(1+choices.index(c), c.name, pad)
            if c.descr:
                entry += "-- {:<}".format(c.descr)
            return entry

        Script.BeginFunc()

        # Pad names to (3 + length of longest name) spaces.
        pad = 3 + max(list(map(lambda c: len(c.name), choices)))
        limit = len(choices)

        # Construct prompt once rather than each time round the query loop.
        prompt = "\n\n## "+title+"\n"+\
            reduce(lambda x,y: x+y, [mkentry(c, pad) for c in choices])+"\n\n> "

        # Keep querying the user until they make a valid selection.
        while True:
            try:
                # Query user.
                sel = int(input(prompt))
                if (sel < 1) or (sel > limit):
                    # Also raised by int() above if not a valid integer.
                    raise ValueError()
                index = sel-1

                # Query again if they chose a disabled option.
                if choices[index].disabled:
                    print("\nOption not available.")
                    continue

                # Return selection.
                elif ret == "index":
                    return index
                elif ret == "reference":
                    return choices[index]
            except ValueError:
                print("!! Error: Expected number in range [1.."
                    +str(limit)+"] inclusive.")

    """
     " Call Menu.Prompt() by auto-generating a list of choice namedtuples
     " from a list of objects that each have a .name and .descr member.
     " Returns a reference to the chosen object in the list.
    """
    def PromptFromList( title, objlist ):
        Script.BeginFunc()
        choices = list(map(lambda o: Menu.Choice(o.name, o.descr), objlist))
        return objlist[Menu.Prompt(title, choices, ret="index")]

"""
 " An online resource that can be downloaded.
"""
class Resource():

    def __init__( self, name, baseurl, filename, dstdir="", md5file=None ):
        self.name = name
        self.filename = filename
        self.baseurl = baseurl+"/"
        self.dstdir = dstdir+"/" # Path is relative to Script.BaseDir()
        self.md5file = md5file

    """
     " Download this resource.
    """
    def Fetch( self, silent=False ):
        Script.BeginFunc()
        dldir = Script.DlDir()+re.sub(r"[:/]", r"_", self.baseurl)+"/"
        Shell.MkDir(dldir)
        dlfile = dldir+self.filename
        expected_md5 = None
        already_fetched = False

        # If this resource has an associated MD5 checksum file, download it and
        # extract this resource's hash.
        if self.md5file:
            Script.Log("Resource has MD5 checksum file")
            md5url = self.baseurl+self.md5file
            md5file = dldir+self.md5file
            if Web.FetchURL(md5url, md5file, self.name+" MD5s", silent=True):
                Script.Log("Successfully fetched MD5 checksum file")
                expected_md5 = Shell.MD5.ExtractHash(md5file, self.filename)
                Script.Log("Expected MD5: "+str(expected_md5))

            # We can skip downloading if the intended local file already exists
            # and its hash matches the expected hash. Note that we always
            # perform a fresh download when Script.TestModeEnabled().
            if os.path.isfile(dlfile) and not Script.TestModeEnabled():
                existing_md5 = Shell.MD5.HashFile(dlfile)
                Script.Log("Existing MD5: "+str(existing_md5))
                if existing_md5 == expected_md5:
                    Script.Log("MD5 hashes match; can skip download")
                    already_fetched = True

        # Fetch this resource.
        if already_fetched:
            print("Already fetched "+self.name, end="")
        else:
            url = self.baseurl+self.filename
            if Web.FetchURL(url, dlfile, self.name, silent=silent):
                Script.Log("Successfully fetched resource")
                if expected_md5:
                    Script.Log("Need to verify integrity")
                    actual_md5 = Shell.MD5.HashFile(dlfile)
                    if actual_md5 == expected_md5:
                        Script.Log("Verification successful")
                    else:
                        Script.Abort("MD5 integrity check failed for " \
                            +self.name+" (file: "+dlfile+"), expected '" \
                            +expected_md5+"' but got '"+actual_md5+"'")

        # Extract this resource if it's an archive, otherwise simply copy it.
        if not Script.TestModeEnabled():
            Shell.MkDir(self.dstdir)
            Shell.Flush()
            print(". Extracting... ", end="")
            Shell.Flush()
            file_was_extracted = Shell.ExtractArchive(dlfile, self.dstdir)
            if not file_was_extracted:
                Script.Log("File was not extracted, copying instead")
                Shell.Copy(dlfile, self.dstdir+self.filename)
            print("Done.")
            Shell.Flush()

    """
     " Passed to MkDict(); represents a new resource that will be keyed in the
     " returned dictionary using the provided key.
    """
    DictTuple = namedtuple("ResourceDictTuple",
        ["key", "name", "filename", "dstdir"])

    """
     " Return a new dictionary of resources associated with a common base URL
     " and MD5 checksum files; tuples is a list of DictTuple namedtuples.
    """
    def MkDict( baseurl, tuples, md5file=None ):
        rdict = {}
        for t in tuples:
            rdict[t.key] = \
                Resource(t.name, baseurl, t.filename, t.dstdir, md5file)
        return rdict

"""
 " Represents a build-script filesystem.
 "
 " Equivalent to class with constructor: (name, script, descr=None, rkeys=None).
 "
 " script is the name of the filesystem build-script associated with this
 " filesystem (for example `busybox', `oe', `uefi', `android', ...), and rkeys
 " is a list of dictionary keys identifying the resources that this filesystem
 " depends on (for example the Android filesystem image dictionary key).
"""
Filesystem = namedtuple("FilesystemTuple", ["name", "script", "descr", "rkeys"])
Filesystem.__new__.__defaults__ = None, None

"""
 " Quick creation of Android resources and filesystems.
"""
class Android():

    """
     " Return a new dictionary containing an Android image resource and Android
     " ramdisk resource, both associated with the same base URL and MD5
     " checksum file. Set ramdkey=None for no ramdisk.
    """
    def MkResDict( baseurl, img, imgkey, ramd="ramdisk.img", ramddir="",
            ramdkey=None, md5file="MD5SUMS" ):

        Script.BeginFunc()
        tuples = [Resource.DictTuple(imgkey, "Android image", img, "")]
        if ramdkey:
            tuples.append(Resource.DictTuple(ramdkey, "Android ramdisk",
                ramd, ramddir))
        return Resource.MkDict(baseurl, tuples, md5file)

    """
     " Return a new dictionary containing an Android filesystem.
     " Set ramdkey=None if this Android filesystem does not require a ramdisk.
    """
    def MkFsDict( release, fskey, imgkey, ramdkey=None ):
        Script.BeginFunc()
        rkeys = [imgkey]
        if ramdkey:
            rkeys.append(ramdkey)
        d = {fskey : Filesystem("Android", "android", release, rkeys)}
        return d


"""
 " Represents a prebuilt configuration.
 "
 " Equivalent to class with constructor: (name, rkeys, descr=None)
 "
 " rkeys is a list of dictionary keys identifying the resources that this
 " prebuilt configuration depends on.
"""
Prebuilt = namedtuple("PrebuiltTuple", ["name", "rkeys", "descr"])
Prebuilt.__new__.__defaults__ = None,

"""
 " Represents a build-script manifest.
 "
 " Equivalent to class with constructor: (manifest, cckeys).
 "
 " manifest is the name of the build-script manifest (for example `ack',
 " `latest'), and cckeys is a list of dictionary keys identifying the compilers
 " required for building against this manifest.
"""
Manifest = namedtuple("ManifestTuple", ["manifest", "cckeys"])

"""
 " Represents a baremetal (pre-OS) environment such as UEFI.
 "
 " Equivalent to class with constructor:
 "      (name, manifest, stubfs, descr=None)
 "
 " manifest points to a manifest object; stubfs is the stub filesystem
 " build-script required by this manifest (e.g. `uefi').
"""
Baremetal = namedtuple("BaremetalTuple",["name", "manifest", "stubfs", "descr"])
Baremetal.__new__.__defaults__ = None,

"""
 " Represents a Linux/Android kernel environment.
 "
 " Equivalent to class with constructor:
 "      (name, manifest, fskeys, descr=None)
 "
 " manifest points to a manifest object; fskeys is a list of dictionary keys
 " identifying the filesystems supported by this kernel.
"""
Kernel = namedtuple("KernelTuple", ["name", "manifest", "fskeys", "descr"])
Kernel.__new__.__defaults__ = None,

"""
 " A platform such as Juno.
"""
class Platform():

    def __init__( self, name, scriptdir, rdicts, fsdicts, kernels, baremetals,
        prebuilts ):

        Script.BeginFunc()
        self.name = name
        self.scriptdir = scriptdir
        self.resources = {}
        if rdicts:
            for rd in rdicts:
                self.resources.update(rd)
        self.filesystems = {}
        if fsdicts:
            for fsd in fsdicts:
                self.filesystems.update(fsd)
        self.kernels = kernels
        self.baremetals = baremetals
        self.prebuilts = prebuilts

    """
     " Query the user for their configuration, then initialise the workspace.
    """
    def QueryConfigAndInitWorkspace( self ):

        """
         " Print a table displaying the user's chosen configuration and ask
         " them to confirm whether they want to initialise their workspace
         " using that configuration; exit the script if they don't.
        """
        def ConfirmOrExit( *args ):

            """
             " Display a prompt asking the user to type 'y'/'Y' to continue,
             " or 'n'/'N' to quit. Any other input will reprompt.
            """
            def confirm( prompt ):
                response = ""
                while not (response == "y" or response == "Y"):
                    response = input("\n"+prompt).strip()
                    if response == "n" or response == "N":
                        exit(1)

            """
             " Get the length of the longest string in a list.
            """
            def maxlen( lst ):
                return max(list(map(lambda x: len(x), lst)))

            Script.BeginFunc()
            titles = ["Workspace", "Platform", "Build"]
            values = [Script.BaseDir(), self.name]

            if type(args[0]) is Prebuilt:
                pb = args[0]
                titles += ["Prebuilt"]
                values += ["Use prebuilt configuration", pb.name]

            elif type(args[0]) is Baremetal:
                bm = args[0]
                titles += ["Environment", "Target"]
                values += ["Build from source", "Baremetal", bm.name]

            elif type(args[0]) is Kernel:
                k, fs = args[0], args[1]
                titles += ["Environment", "Kernel", "Filesystem"]
                values += ["Build from source", "Linux/Android", k.name]
                fsname = fs.name
                if fs.descr:
                    fsname += " "+fs.descr
                values += [fsname]

            # Pad each column to the longest string in that column.
            titlepad, valuepad = maxlen(titles), maxlen(values)
            wall = "    +-" + (titlepad*"-") + "-+-" + (valuepad*"-") + "-+"

            print("\nYour chosen configuration is as follows:\n")
            print(wall)
            for row in zip(titles, values):
                print("    | {:<{:}} | {:<{:}} |".format(row[0], titlepad,
                    row[1], valuepad))
            print(wall)

            confirm("Proceed with this configuration? [y/n] > ")

            # Generate list of all files and directories in Script.BaseDir().
            filelist = []
            for (dirpath, dirnames, filenames) in os.walk(Script.BaseDir()):
                for d in dirnames:
                    # Just doing d+="/" doesn't seem to work; scoped copy?
                    dirnames[dirnames.index(d)]+="/"
                filelist.extend(dirnames)
                filelist.extend(filenames)
                filelist.sort()
                break  # Break at root level (i.e. don't recursively walk).

            # Remove script, download directory, and repo directory from list.
            for f in \
            [
                __file__.rsplit("/")[-1],
                __file__.rsplit("\\")[-1],
                Script.DlDir(),
                ".repo/",
            ]:
                if f in filelist:
                    filelist.remove(f)

            if filelist:
                confirm("## [WARNING] Expected empty workspace but found the "
                    "following files & folders:\n"\
                         +reduce(lambda x,y: x+y,
                            map(lambda f: "\n  - "+f, filelist)) + "\n\nDelete "
                                "these files & folders and proceed? [y/n] > ")
                for path in filelist:
                    rmfunc = Shell.RmDir if path.endswith("/") else Shell.RmFile
                    rmfunc(path)
            print()  # For newline

        # Construct choice for "Build from source".
        descr, disabled = None, False
        if not (self.baremetals or self.kernels):
            descr. disabled = "<DISABLED> Not supported by platform", True
        elif not Script.RunningOnLinux():
            descr, disabled = "<DISABLED> Requires Linux host", True
        build_from_source = Menu.Choice("Build from source", descr, disabled)

        # Construct choice for "Use prebuilt configuration".
        descr, disabled = None, False
        if not self.prebuilts:
            descr, disabled = "<DISABLED> Not provided for platform", True
        use_prebuilt = Menu.Choice("Use prebuilt config", descr, disabled)

        # Prompt for build type.
        btype = Menu.Prompt("Please specify whether you want to:",
            [build_from_source, use_prebuilt], ret="reference")

        # Prompt for prebuilt configuration.
        if btype is use_prebuilt:
            pb = Menu.PromptFromList("Please select a prebuilt config:",
                self.prebuilts)
            ConfirmOrExit(pb)
            for rkey in pb.rkeys:
                self.resources[rkey].Fetch()
            print("\nWorkspace initialised.")

        else:

            """
             " Use `dpkg-query' to check whether an apt package is installed.
            """
            def gotdep( d ):
                devnull = subprocess.DEVNULL
                return 0 == subprocess.call(["dpkg-query", "-s", d],
                    stdout=devnull, stderr=devnull)

            print("\nChecking dependencies... ", end="")
            missing = []
            for d in SCRIPT_DEPS:
                if not gotdep(d):
                    Script.Log("Missing "+d)
                    missing.append(d)
            if missing:
                missing.sort()
                Script.Abort("The following packages are missing:\n"\
                    +reduce(lambda x,y: x+y, \
                        map(lambda m: "\n  - "+m,missing))+"\n\nPlease install"\
                           "install these missing packages using the following "\
                           "command:\n\n$ sudo apt-get install"+reduce(lambda\
                                x,y: x+y, map(lambda m: " "+m, missing)))
            print("OK.")

            # Construct choice for "Linux/Android".
            descr, disabled = None, False
            if not self.kernels:
                descr, disabled = "<DISABLED> Not supported by platform", True
            env_linux = Menu.Choice("Linux/Android", descr, disabled)

            # Construct choice for "Baremetal".
            descr, disabled = None, False
            if not self.baremetals:
                descr, disabled = "<DISABLED> Not supported by platform", True
            env_baremetal = Menu.Choice("Baremetal", descr, disabled)

            # Prompt for environment.
            env = Menu.Prompt("Please select an environment:",
                [env_linux, env_baremetal], ret="reference")

            manifest, rkeys, fsscript = None, None, None

            #  Prompt for baremetal target.
            if env is env_baremetal:
                bm = Menu.PromptFromList("Please select a target:",
                    self.baremetals)
                ConfirmOrExit(bm)
                manifest = bm.manifest.manifest
                rkeys = bm.manifest.cckeys
                fsscript = bm.stubfs

            # Prompt for kernel and filesystem.
            else:
                k = Menu.PromptFromList("Please select a kernel:", self.kernels)
                filesystems = list(map(lambda fskey: self.filesystems[fskey],
                    k.fskeys))
                fs = Menu.PromptFromList("Please select a filesystem:",
                    filesystems)
                ConfirmOrExit(k, fs)
                manifest = k.manifest.manifest
                rkeys = k.manifest.cckeys
                if fs.rkeys:
                    rkeys += fs.rkeys
                fsscript = fs.script

            # Manifest workspace.
            for rkey in rkeys:
                self.resources[rkey].Fetch()
            Web.RepoSync(manifest)
            for (dirpath, dirnames, filenames) in \
                os.walk(Script.BaseDir()+"/build-scripts/platforms/"):
                for d in dirnames:
                    if not (d == "common" or d == self.scriptdir):
                        Shell.RmDir("build-scripts/platforms/"+d)
                break  # Break at root level (i.e. don't recursively walk)
            for (dirpath, dirnames, filenames) in \
                os.walk(Script.BaseDir()+"/build-scripts/filesystems/"):
                for f in filenames:
                    if not (f == fsscript):
                        Shell.RmFile("build-scripts/filesystems/"+f)
                break  # Break at root level (i.e. don't recursively walk)

            print("\nWorkspace initialised.")
            print("\nTo build:")
            print("\n    chmod a+x <workspace>/build-scripts/build-all.sh")
            print("    <workspace>/build-scripts/build-all.sh all")
            print("\nResulting binaries will be placed in:")
            print("\n    <workspace>/output/"+self.scriptdir+"/"+self.scriptdir
                +"-"+fsscript+"/")

        print("\nFor more information, tutorials, FAQs, and discussions, please see here:")
        print("\n    https://www.community.arm.com/tools/dev-platforms/")
        print("\nThank you for using the Linaro ARM Platforms workspace script.")
        return 0

"""
 " Shared cross compiler resources.
"""
class Compiler():

    __aa32 = None
    __aa64 = None

    def AArch32():
        if not Compiler.__aa32:
            fn = "gcc-linaro-"+AARCH32_COMPILER_VSN+"-x86_64_arm-linux-gnueabihf.tar.xz"
            Compiler.__aa32 = Resource\
            (
                name="AArch32 compiler",
                baseurl="http://releases.linaro.org/components/toolchain/"\
                    +"binaries/"+COMPILER_RELEASE+"/arm-linux-gnueabihf/",
                filename=fn,
                dstdir="tools/gcc/",
                md5file=fn+".asc",
            )
        return Compiler.__aa32

    def AArch64():
        if not Compiler.__aa64:
            fn = "gcc-linaro-"+AARCH64_COMPILER_VSN+"-x86_64_aarch64-linux-gnu.tar.xz"
            Compiler.__aa64 = Resource\
            (
                name="AArch64 compiler",
                baseurl="http://releases.linaro.org/components/toolchain/"\
                    +"binaries/"+COMPILER_RELEASE+"/aarch64-linux-gnu/",
                filename=fn,
                dstdir="tools/gcc/",
                md5file=fn+".asc",
            )
        return Compiler.__aa64




"""
 " Create a new platform representing the ARM Juno development board
 " where EL3 remains AArch64.
"""
def Juno64():
    nameprefix = "Juno 64-bit "
    oeminif="lt-vexpress64-openembedded_minimal-armv8-gcc-"\
        +JUNO64_OE_MINI_VSN+".img.gz"
    oelampf="lt-vexpress64-openembedded_lamp-armv8-gcc-"\
        +JUNO64_OE_LAMP_VSN+".img.gz"    
    return Platform\
    (
        name="[64-bit] Juno",
        scriptdir="juno",
        rdicts=\
        [
            Resource.MkDict\
            (
                md5file="MD5SUMS",
                baseurl="http://releases.linaro.org/members/arm/android/juno/"
                    +JUNO64_ANDROID_RELEASE,
                tuples=\
                [
                    Resource.DictTuple\
                    (
                        key=KEY_ANDROID_IMAGE,
                        name=nameprefix+"Android filesystem image",
                        filename="juno.img.xz",
                        dstdir="",
                    ),
                    Resource.DictTuple\
                    (
                        key=KEY_ANDROID_RAMDISK,
                        name=nameprefix+"Android ramdisk image",
                        filename="ramdisk.img",
                        dstdir="prebuilts/android/juno/",
                    ),
                ],
            ),
            {
                KEY_OE_MINI_IMAGE: Resource\
                (
                    name=nameprefix+"OpenEmbedded Minimal image",
                    baseurl="https://releases.linaro.org/openembedded/juno-lsk/"
                        +JUNO64_OE_RELEASE,
                    filename=oeminif,
                    dstdir="",
                    md5file=oeminif+".md5",
                ),
                KEY_OE_LAMP_IMAGE: Resource\
                (
                    name=nameprefix+"OpenEmbedded LAMP image",
                    baseurl="https://releases.linaro.org/openembedded/juno-lsk/"
                        +JUNO64_OE_RELEASE,
                    filename=oelampf,
                    dstdir="",
                    md5file=oelampf+".md5",
                ),            
            },    
            Resource.MkDict\
            (
                md5file="MD5SUMS",
                baseurl="http://releases.linaro.org/members/arm/platforms/"
                    +JUNO64_PREBUILT_RELEASE,
                tuples=\
                [
                    Resource.DictTuple\
                    (
                        key=KEY_BUSYBOX_LATEST_PREBUILT,
                        name="juno-latest-busybox-uboot.zip",
                        filename="juno-latest-busybox-uboot.zip",
                        dstdir="juno-latest-busybox-uboot",
                    ),
                    Resource.DictTuple\
                    (
                        key=KEY_OE_PREBUILT,
                        name="juno-latest-oe-uboot.zip",
                        filename="juno-latest-oe-uboot.zip",
                        dstdir="juno-latest-oe-uboot",
                    ),
                    Resource.DictTuple\
                    (
                        key=KEY_ANDROID_PREBUILT,
                        name="juno-ack-android-uboot.zip",
                        filename="juno-ack-android-uboot.zip",
                        dstdir="juno-ack-android-uboot",
                    ),
                    Resource.DictTuple\
                    (
                        key=KEY_UEFI_PREBUILT,
                        name="juno-uefi.zip",
                        filename="juno-uefi.zip",
                        dstdir="juno-uefi",
                    ),
                ],
            ),
            {
                KEY_AARCH32_COMPILER: Compiler.AArch32(),
                KEY_AARCH64_COMPILER: Compiler.AArch64(),
            },
        ],
        fsdicts=\
        [
            {
                KEY_BUSYBOX_FS : Filesystem\
                (
                    name="BusyBox",
                    script="busybox",
                    descr="Built from source",
                ),
                KEY_ANDROID_FS : Filesystem\
                (
                    name="Android",
                    script="android",
                    descr="O "+JUNO64_ANDROID_RELEASE,
                    rkeys=\
                    [
                        KEY_ANDROID_IMAGE,
                        KEY_ANDROID_RAMDISK,
                    ],
                ),
                KEY_OE_MINI_FS : Filesystem\
                (
                    name="OpenEmbedded Minimal",
                    script="oe",
                    descr=JUNO64_OE_RELEASE,
                    rkeys=[ KEY_OE_MINI_IMAGE ],
                ),
                KEY_OE_LAMP_FS : Filesystem\
                (
                    name="OpenEmbedded LAMP",
                    script="oe",
                    descr=JUNO64_OE_RELEASE,
                    rkeys=[ KEY_OE_LAMP_IMAGE ],
                ),
            },
        ],
        kernels=\
        [
            Kernel\
            (
                name="ack-4.9-armlt",
                descr="Supports Android",
                manifest=Manifest\
                (
                    manifest="ack",
                    cckeys=\
                    [
                        KEY_AARCH32_COMPILER,
                        KEY_AARCH64_COMPILER,
                    ],
                ),
                fskeys=\
                [
                    KEY_BUSYBOX_FS,
                    KEY_OE_MINI_FS,
                    KEY_OE_LAMP_FS,
                    KEY_ANDROID_FS,
                ],
            ),
            Kernel\
            (
                name="latest-armlt",
                manifest=Manifest\
                (
                    manifest="latest",
                    cckeys=\
                    [
                        KEY_AARCH32_COMPILER,
                        KEY_AARCH64_COMPILER,
                    ],
                ),
                fskeys=\
                [
                    KEY_BUSYBOX_FS,
                    KEY_OE_MINI_FS,
                    KEY_OE_LAMP_FS,
                ],
            ),               
        ],
        baremetals=\
        [
            Baremetal\
            (
                name="EDK2 (UEFI)",
                stubfs="uefi",
                manifest=Manifest\
                (
                    manifest="uefi",
                    cckeys=\
                    [
                        KEY_AARCH32_COMPILER,
                        KEY_AARCH64_COMPILER,
                    ],
                ),
            ),
        ],
        prebuilts=\
        [
            Prebuilt\
            (
                name="EDK2 (UEFI)",
                rkeys=[ KEY_UEFI_PREBUILT ],
            ),
            Prebuilt\
            (
                name="latest-armlt + BusyBox",
                rkeys=[ KEY_BUSYBOX_LATEST_PREBUILT ],
            ),
            Prebuilt\
            (
                name="latest-armlt + OpenEmbedded Minimal "+JUNO64_OE_RELEASE,
                rkeys=\
                [
                    KEY_OE_PREBUILT,
                    KEY_OE_MINI_IMAGE,
                ],
            ),
            Prebuilt\
            (
                name="latest-armlt + OpenEmbedded LAMP "+JUNO64_OE_RELEASE,
                rkeys=\
                [
                    KEY_OE_PREBUILT,
                    KEY_OE_LAMP_IMAGE,
                ],
            ),
            Prebuilt\
            (
                name="ack-4.9-armlt + Android O "+JUNO64_ANDROID_RELEASE,
                rkeys=\
                [
                    KEY_ANDROID_PREBUILT,
                    KEY_ANDROID_IMAGE,
                ],
            ),
            Prebuilt\
            (
                name="(( All available prebuilt configurations ))",
                rkeys=\
                [
                    KEY_UEFI_PREBUILT,
                    KEY_BUSYBOX_LATEST_PREBUILT,
                    KEY_OE_PREBUILT,
                    KEY_ANDROID_PREBUILT,
                    KEY_OE_MINI_IMAGE,
                    KEY_OE_LAMP_IMAGE,
                    KEY_ANDROID_IMAGE,
                ],
            ),
        ]
    )


"""
 " Create a new platform representing the ARM Juno development board
 " where EL3 is warm-reset into AArch32.
"""
def Juno32():
    nameprefix = "Juno 32-bit "
    return Platform\
    (
        name="[32-bit] Juno",
        scriptdir="juno32",
        rdicts=\
        [
            Resource.MkDict\
            (
                md5file=None, # OE ALIP has no MD5 checksum file.
                baseurl="http://releases.linaro.org/openembedded/vexpress-lsk/"
                    +JUNO32_OE_RELEASE,
                tuples=\
                [
                    Resource.DictTuple
                    (
                        key=KEY_OE_ALIP_IMAGE,
                        name=nameprefix+"OpenEmbedded ALIP image",
                        filename="lsk-vexpress-openembedded_alip-armv7a-gcc-"
                            +JUNO32_OE_ALIP_VSN+".img.gz",
                        dstdir="",
                    ),
                ],
            ),
            Resource.MkDict\
            (
                md5file="MD5SUMS",
                baseurl="http://releases.linaro.org/members/arm/"\
                    +"platforms/"+JUNO32_PREBUILT_RELEASE+"/",
                tuples=\
                [
                    Resource.DictTuple
                    (
                        key=KEY_BUSYBOX_LATEST_PREBUILT,
                        name="juno32-latest-busybox-uboot.zip",
                        filename="juno32-latest-busybox-uboot.zip",
                        dstdir="juno32-latest-busybox-uboot",
                    ),
                    Resource.DictTuple
                    (
                        key=KEY_BUSYBOX_ACK_PREBUILT,
                        name="juno32-ack-busybox-uboot.zip",
                        filename="juno32-ack-busybox-uboot.zip",
                        dstdir="juno32-ack-busybox-uboot",
                    ),
                    Resource.DictTuple
                    (
                        key=KEY_OE_PREBUILT,
                        name="juno32-latest-oe-uboot.zip",
                        filename="juno32-latest-oe-uboot.zip",
                        dstdir="juno32-latest-oe-uboot",
                    ),
                ],
            ),
            {
                KEY_AARCH32_COMPILER: Compiler.AArch32(),
                KEY_AARCH64_COMPILER: Compiler.AArch64(),
            },
        ],
        fsdicts=\
        [
            {
                KEY_BUSYBOX_FS : Filesystem\
                (
                    name="BusyBox",
                    script="busybox",
                    descr="Built from source",
                ),
                KEY_OE_ALIP_FS : Filesystem\
                (
                    name="OpenEmbedded ALIP",
                    script="oe",
                    descr=JUNO32_OE_RELEASE,
                    rkeys=[ KEY_OE_ALIP_IMAGE ],
                ),
            },
        ],
        kernels=\
        [
            Kernel\
            (
                name="ack-4.9-armlt",
                manifest=Manifest\
                (
                    manifest="ack",
                    cckeys=\
                    [
                        KEY_AARCH32_COMPILER,
                        KEY_AARCH64_COMPILER,
                    ],
                ),
                fskeys=\
                [
                    KEY_BUSYBOX_FS,
                    KEY_OE_ALIP_FS,
                ],
            ),
            Kernel\
            (
                name="latest-armlt",
                manifest=Manifest\
                (
                    manifest="latest",
                    cckeys=\
                    [
                        KEY_AARCH32_COMPILER,
                        KEY_AARCH64_COMPILER,
                    ],
                ),
                fskeys=\
                [
                    KEY_BUSYBOX_FS,
                    KEY_OE_ALIP_FS,
                ],
            ), 
        ],
        baremetals=None,
        prebuilts=\
        [
            Prebuilt\
            (
                name="ack-4.9-armlt + BusyBox",
                rkeys=[ KEY_BUSYBOX_ACK_PREBUILT ],
            ),
            Prebuilt\
            (
                name="latest-armlt + BusyBox",
                rkeys=[ KEY_BUSYBOX_LATEST_PREBUILT ],
            ),
            Prebuilt\
            (
                name="latest-armlt + OpenEmbedded ALIP "+JUNO32_OE_RELEASE,
                rkeys=\
                [
                    KEY_OE_PREBUILT,
                    KEY_OE_ALIP_IMAGE,
                ],
            ),
        ],
    )


"""
 " Create a new platform representing the the AEMv8-A Base Platform
 " where EL3 is configured to be AArch64.
"""
def FVP64():
    nameprefix = "FVP 64-bit "
    return Platform\
    (
        name="[64-bit] AEMv8-A Base Platform FVP",
        scriptdir="fvp",
        rdicts=\
        [
            Resource.MkDict\
            (
                md5file="MD5SUMS",
                baseurl="http://releases.linaro.org/android/reference-lcr/fvp/"
                    +FVP64_ANDROID_RELEASE,
                tuples=\
                [
                    Resource.DictTuple\
                    (
                        key=KEY_ANDROID_IMAGE,
                        name=nameprefix+"Android filesystem image",
                        filename="fvp.img.bz2",
                        dstdir="",
                    ),
                    Resource.DictTuple\
                    (
                        key=KEY_ANDROID_RAMDISK,
                        name=nameprefix+"Android ramdisk image",
                        filename="ramdisk.img",
                        dstdir="prebuilts/android/fvp/",
                    ),
                ],
            ),
            Resource.MkDict\
            (
                md5file="MD5SUMS.txt",
                baseurl="https://releases.linaro.org/openembedded/juno-lsk/"
                    +FVP64_OE_RELEASE,
                tuples=\
                [
                    Resource.DictTuple\
                    (
                        key=KEY_OE_MINI_IMAGE,
                        name=nameprefix+"OpenEmbedded Minimal image",
                        filename="lt-vexpress64-openembedded_minimal-armv8-gcc-"
                            +FVP64_OE_MINI_VSN+".img.gz",
                        dstdir="",
                    ),
                    Resource.DictTuple\
                    (
                        key=KEY_OE_LAMP_IMAGE,
                        name=nameprefix+"OpenEmbedded LAMP image",
                        filename="lt-vexpress64-openembedded_lamp-armv8-gcc-"
                            +FVP64_OE_LAMP_VSN+".img.gz",
                        dstdir="",
                    ),
                ],
            ),
            Resource.MkDict\
            (
                md5file="MD5SUMS",
                baseurl="http://releases.linaro.org/members/arm/platforms/"
                    +FVP64_PREBUILT_RELEASE,
                tuples=\
                [
                    Resource.DictTuple\
                    (
                        key=KEY_BUSYBOX_LATEST_PREBUILT,
                        name="fvp-latest-busybox-uboot.zip",
                        filename="fvp-latest-busybox-uboot.zip",
                        dstdir="fvp-latest-busybox-uboot",
                    ),
                    Resource.DictTuple\
                    (
                        key=KEY_BUSYBOX_ACK_PREBUILT,
                        name="fvp-ack-busybox-uboot.zip",
                        filename="fvp-ack-busybox-uboot.zip",
                        dstdir="fvp-ack-busybox-uboot",
                    ),
                    Resource.DictTuple\
                    (
                        key=KEY_OE_PREBUILT,
                        name="fvp-latest-oe-uboot.zip",
                        filename="fvp-latest-oe-uboot.zip",
                        dstdir="fvp-latest-oe-uboot",
                    ),
                    Resource.DictTuple\
                    (
                        key=KEY_ANDROID_PREBUILT,
                        name="fvp-ack-android-uboot.zip",
                        filename="fvp-ack-android-uboot.zip",
                        dstdir="fvp-ack-android-uboot",
                    ),
                    Resource.DictTuple\
                    (
                        key=KEY_UEFI_PREBUILT,
                        name="fvp-uefi.zip",
                        filename="fvp-uefi.zip",
                        dstdir="fvp-uefi",
                    ),
                    Resource.DictTuple\
                    (
                        key=KEY_OE_DEBUG_PREBUILT,
                        name="fvp-latest-oe-debug.tar.xz",
                        filename="fvp-latest-oe-debug.tar.xz",
                        dstdir="fvp-latest-oe-debug",
                    ),
                    Resource.DictTuple\
                    (
                        key=KEY_ANDROID_DEBUG_PREBUILT,
                        name="fvp-ack-android-debug.tar.xz",
                        filename="fvp-ack-android-debug.tar.xz",
                        dstdir="fvp-ack-android-debug",
                    ),
                ],
            ),
            {
                KEY_AARCH32_COMPILER: Compiler.AArch32(),
                KEY_AARCH64_COMPILER: Compiler.AArch64(),
            },
        ],
        fsdicts=\
        [
            {
                KEY_BUSYBOX_FS : Filesystem\
                (
                    name="BusyBox",
                    script="busybox",
                    descr="Built from source",
                ),
                KEY_ANDROID_FS : Filesystem\
                (
                    name="Android",
                    script="android",
                    descr="N "+FVP64_ANDROID_RELEASE,
                    rkeys=\
                    [
                        KEY_ANDROID_IMAGE,
                        KEY_ANDROID_RAMDISK,
                    ],
                ),
                KEY_OE_MINI_FS : Filesystem\
                (
                    name="OpenEmbedded Minimal",
                    script="oe",
                    descr=FVP64_OE_RELEASE,
                    rkeys=[ KEY_OE_MINI_IMAGE ],
                ),
                KEY_OE_LAMP_FS : Filesystem\
                (
                    name="OpenEmbedded LAMP",
                    script="oe",
                    descr=FVP64_OE_RELEASE,
                    rkeys=[ KEY_OE_LAMP_IMAGE ],
                ),
            },
        ],
        kernels=\
        [
            Kernel\
            (
                name="ack-4.9-armlt",
                descr="Supports Android",
                manifest=Manifest\
                (
                    manifest="ack",
                    cckeys=\
                    [
                        KEY_AARCH32_COMPILER,
                        KEY_AARCH64_COMPILER,
                    ],
                ),
                fskeys=\
                [
                    KEY_BUSYBOX_FS,
                    KEY_OE_MINI_FS,
                    KEY_OE_LAMP_FS,
                    KEY_ANDROID_FS,
                ],
            ),
            Kernel\
            (
                name="latest-armlt",
                manifest=Manifest\
                (
                    manifest="latest",
                    cckeys=\
                    [
                        KEY_AARCH32_COMPILER,
                        KEY_AARCH64_COMPILER,
                    ],
                ),
                fskeys=\
                [
                    KEY_BUSYBOX_FS,
                    KEY_OE_MINI_FS,
                    KEY_OE_LAMP_FS,
                ],
            ),               
        ],
        baremetals=\
        [
            Baremetal\
            (
                name="EDK2 (UEFI)",
                stubfs="uefi",
                manifest=Manifest\
                (
                    manifest="uefi",
                    cckeys=\
                    [
                        KEY_AARCH32_COMPILER,
                        KEY_AARCH64_COMPILER,
                    ],
                ),
            ),
        ],
        prebuilts=\
        [
            Prebuilt\
            (
                name="EDK2 (UEFI)",
                rkeys=[ KEY_UEFI_PREBUILT ],
            ),
            Prebuilt\
            (
                name="ack-4.9-armlt + BusyBox",
                rkeys=[ KEY_BUSYBOX_ACK_PREBUILT ],
            ),
            Prebuilt\
            (
                name="latest-armlt + BusyBox",
                rkeys=[ KEY_BUSYBOX_LATEST_PREBUILT ],
            ),
            Prebuilt\
            (
                name="latest-armlt + OpenEmbedded Minimal "+FVP64_OE_RELEASE,
                rkeys=\
                [
                    KEY_OE_PREBUILT,
                    KEY_OE_MINI_IMAGE,
                ],
            ),
            Prebuilt\
            (
                name="latest-armlt + OpenEmbedded LAMP "+FVP64_OE_RELEASE,
                rkeys=\
                [
                    KEY_OE_PREBUILT,
                    KEY_OE_LAMP_IMAGE,
                ],
            ),
            Prebuilt\
            (
                name="ack-4.9-armlt + Android N "+FVP64_ANDROID_RELEASE,
                rkeys=\
                [
                    KEY_ANDROID_PREBUILT,
                    KEY_ANDROID_IMAGE,
                ],
            ),
            Prebuilt\
            (
                name="[DEBUG] latest-armlt + OpenEmbedded Minimal "
                    +FVP64_OE_RELEASE,
                rkeys=\
                [
                    KEY_OE_PREBUILT,
                    KEY_OE_DEBUG_PREBUILT,
                    KEY_OE_MINI_IMAGE,
                ],
            ),
            Prebuilt\
            (
                name="[DEBUG] latest-armlt + OpenEmbedded LAMP "
                    +FVP64_OE_RELEASE,
                rkeys=\
                [
                    KEY_OE_PREBUILT,
                    KEY_OE_DEBUG_PREBUILT,
                    KEY_OE_LAMP_IMAGE,
                ],
            ),
            Prebuilt\
            (
                name="[DEBUG] ack-4.9-armlt + Android N "+FVP64_ANDROID_RELEASE,
                rkeys=\
                [
                    KEY_ANDROID_PREBUILT,
                    KEY_ANDROID_DEBUG_PREBUILT,
                    KEY_ANDROID_IMAGE,
                ],
            ),
            Prebuilt\
            (
                name="(( All available prebuilt configurations ))",
                rkeys=\
                [
                    KEY_UEFI_PREBUILT,
                    KEY_BUSYBOX_LATEST_PREBUILT,
                    KEY_OE_PREBUILT,
                    KEY_OE_DEBUG_PREBUILT,
                    KEY_ANDROID_PREBUILT,
                    KEY_ANDROID_DEBUG_PREBUILT,
                    KEY_OE_MINI_IMAGE,
                    KEY_OE_LAMP_IMAGE,
                    KEY_ANDROID_IMAGE,
                ],
            ),
        ]
    )


"""
 " Create a new platform representing the the AEMv8-A Base Platform
 " where EL3 is configured to be AArch32.
"""
def FVP32():
    nameprefix = "FVP 32-bit "
    return Platform\
    (
        name="[32-bit] AEMv8-A Base Platform FVP",
        scriptdir="fvp32",
        rdicts=\
        [
            Resource.MkDict\
            (
                md5file=None, # OE ALIP has no MD5 checksum file.
                baseurl="http://releases.linaro.org/openembedded/vexpress-lsk/"
                    +FVP32_OE_RELEASE,
                tuples=\
                [
                    Resource.DictTuple\
                    (
                        key=KEY_OE_ALIP_IMAGE,
                        name="FVP AArch32 OpenEmbedded ALIP image",
                        filename="lsk-vexpress-openembedded_alip-armv7a-gcc-"
                            +FVP32_OE_ALIP_VSN+".img.gz",
                        dstdir="",
                    ),
                ],
            ),
            Resource.MkDict\
            (
                md5file="MD5SUMS",
                baseurl="http://releases.linaro.org/members/arm/"\
                    +"platforms/"+FVP32_PREBUILT_RELEASE+"/",
                tuples=\
                [
                    Resource.DictTuple\
                    (
                        key=KEY_BUSYBOX_LATEST_PREBUILT,
                        name="fvp32-latest-busybox-uboot.zip",
                        filename="fvp32-latest-busybox-uboot.zip",
                        dstdir="fvp32-latest-busybox-uboot",
                    ),
                    Resource.DictTuple\
                    (
                        key=KEY_BUSYBOX_ACK_PREBUILT,
                        name="fvp32-ack-busybox-uboot.zip",
                        filename="fvp32-ack-busybox-uboot.zip",
                        dstdir="fvp32-ack-busybox-uboot",
                    ),
                    Resource.DictTuple\
                    (
                        key=KEY_OE_PREBUILT,
                        name="fvp32-latest-oe-uboot.zip",
                        filename="fvp32-latest-oe-uboot.zip",
                        dstdir="fvp32-latest-oe-uboot",
                    ),
                ],
            ),
            {
                KEY_AARCH32_COMPILER: Compiler.AArch32(),
            },
        ],
        fsdicts=\
        [
            {
                KEY_BUSYBOX_FS : Filesystem\
                (
                    name="BusyBox",
                    script="busybox",
                    descr="Built from source",
                ),
                KEY_OE_ALIP_FS : Filesystem\
                (
                    name="OpenEmbedded ALIP",
                    script="oe",
                    descr=FVP32_OE_RELEASE,
                    rkeys=[ KEY_OE_ALIP_IMAGE ],
                ),
            },
        ],
        kernels=\
        [
            Kernel\
            (
                name="ack-4.9-armlt",
                manifest=Manifest\
                (
                    manifest="ack",
                    cckeys=\
                    [
                        KEY_AARCH32_COMPILER,
                    ],
                ),
                fskeys=\
                [
                    KEY_BUSYBOX_FS,
                    KEY_OE_ALIP_FS,
                ],
            ),
            Kernel\
            (
                name="latest-armlt",
                manifest=Manifest\
                (
                    manifest="latest",
                    cckeys=\
                    [
                        KEY_AARCH32_COMPILER,
                    ],
                ),
                fskeys=\
                [
                    KEY_BUSYBOX_FS,
                    KEY_OE_ALIP_FS,
                ],
            ), 
        ],
        baremetals=None,
        prebuilts=\
        [
            Prebuilt\
            (
                name="ack-4.9-armlt + BusyBox",
                rkeys=[ KEY_BUSYBOX_ACK_PREBUILT ],
            ),
            Prebuilt\
            (
                name="latest-armlt + BusyBox",
                rkeys=[ KEY_BUSYBOX_LATEST_PREBUILT ],
            ),
            Prebuilt\
            (
                name="latest-armlt + OpenEmbedded ALIP "+FVP32_OE_RELEASE,
                rkeys=\
                [
                    KEY_OE_PREBUILT,
                    KEY_OE_ALIP_IMAGE,
                ],
            ),
        ],
    )


"""
 " The ARMv8 Foundation Model uses the same deliverables as the 64-bit AEMv8-A
 " Base Platform FVP.
"""
def FoundationModel():
    p = FVP64()
    p.name = "[64-bit] ARMv8 Foundation Model FVP"
    return p


"""
 " Create new platform representing the TC2 development board.
"""
def TC2():
    nameprefix = "TC2 "
    return Platform\
    (
        name="[32-bit] V2P-CA15x2_CA7x3 (TC2)",
        scriptdir="tc2",      
        rdicts=\
        [
            Resource.MkDict\
            (
                md5file="MD5SUMS",
                baseurl="http://releases.linaro.org/android/reference-lcr/"
                    +"vexpress/"+TC2_ANDROID_RELEASE,
                tuples=\
                [
                    Resource.DictTuple\
                    (
                        key=KEY_ANDROID_IMAGE,
                        name="TC2 Android filesystem image",
                        filename="vexpress.img.bz2",
                        dstdir="",
                    ),
                    Resource.DictTuple\
                    (
                        key=KEY_ANDROID_RAMDISK,
                        name="TC2 Android ramdisk image",
                        filename="ramdisk.img",
                        dstdir="prebuilts/android/tc2/",
                    ),
                ],
            ),
            Resource.MkDict\
            (
                md5file=None, # OE ALIP has no MD5 checksum file.
                baseurl="http://releases.linaro.org/openembedded/vexpress-lsk/"
                    +TC2_OE_RELEASE,
                tuples=\
                [
                    Resource.DictTuple\
                    (
                        key=KEY_OE_ALIP_IMAGE,
                        name="TC2 OpenEmbedded ALIP image",
                        filename="lsk-vexpress-openembedded_alip-armv7a-gcc-"
                            +TC2_OE_ALIP_VSN+".img.gz",
                        dstdir="",
                    ),
                ],
            ),
            {
                KEY_AARCH32_COMPILER: Compiler.AArch32(),
            },
        ],
        fsdicts=\
        [
            {
                KEY_BUSYBOX_FS : Filesystem\
                (
                    name="BusyBox",
                    script="busybox",
                    descr="Built from source",
                ),
                KEY_ANDROID_FS : Filesystem\
                (
                    name="Android",
                    script="android",
                    descr="M "+TC2_ANDROID_RELEASE,
                    rkeys=\
                    [
                        KEY_ANDROID_IMAGE,
                        KEY_ANDROID_RAMDISK,
                    ],
                ),
                KEY_OE_ALIP_FS : Filesystem\
                (
                    name="OpenEmbedded ALIP",
                    script="oe",
                    descr=TC2_OE_RELEASE,
                    rkeys=[KEY_OE_ALIP_IMAGE],
                ),
            },
        ],
        kernels=\
        [
            Kernel\
            (
                name="ack-4.9-armlt",
                descr="Supports Android",
                manifest=Manifest\
                (
                    manifest="ack",
                    cckeys=\
                    [
                        KEY_AARCH32_COMPILER,
                    ],
                ),
                fskeys=\
                [
                    KEY_BUSYBOX_FS,
                    KEY_OE_ALIP_FS,
                    KEY_ANDROID_FS,
                ],
            ),
            Kernel\
            (
                name="latest-armlt",
                manifest=Manifest\
                (
                    manifest="latest",
                    cckeys=\
                    [
                        KEY_AARCH32_COMPILER,
                    ],
                ),
                fskeys=\
                [
                    KEY_BUSYBOX_FS,
                    KEY_OE_ALIP_FS,
                ],
            ),               
        ],
        baremetals=\
        [
            Baremetal\
            (
                name="EDK2 (UEFI)",
                stubfs="uefi",
                manifest=Manifest\
                (
                    manifest="uefi",
                    cckeys=\
                    [
                        KEY_AARCH32_COMPILER,
                    ],
                ),
            ),
        ],
        prebuilts=None,
    )


if __name__ == "__main__":

    Script.Init()

    platforms = \
    [
        Juno64(),
        Juno32(),
        FVP64(),
        FVP32(),
        FoundationModel(),
        TC2(),
    ]

    if Script.TestModeEnabled():
        for p in platforms:
            print("*"*20)
            print("Platform: "+p.name)
            if p.resources:
                print("-"*10)
                keys = list(p.resources.keys())
                keys.sort()
                for rkey in keys:
                    r = p.resources[rkey]
                    print("Key:      0x"+format(rkey,"x"))
                    print("Resource: "+str(r.name))
                    print("BaseURL:  "+str(r.baseurl))
                    print("File:     "+str(r.filename))
                    print("Dir:      <workspace>/"+str(r.dstdir))
                    print("MD5File:  "+str(r.md5file))
                    r.Fetch()
                    print("\n"+("-"*10))

    else:

        lst = []
        for p in platforms:
            name, descr, disabled = p.name, None, False
            if (p.baremetals or p.kernels) and not Script.RunningOnLinux() and \
                not p.prebuilts:
                    descr, disabled = "<DISABLED> Can only build from source"\
                        " which requires Linux host", True
            lst.append(Menu.Choice(name, descr, disabled))

        index = Menu.Prompt("Please specify your platform:", lst, ret="index")
        platforms[index].QueryConfigAndInitWorkspace()
